/**
  *  \file gfx/gfx.h
  *  \brief Graphics Kernel Interface
  *
  *  Prototypes and declarations for graphics kernel (higher-level
  *  API wrapped around SDL).
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_GFX_H
#define PCC_V2_GFX_GFX_H

#include "gfx/gfxtypes.h"
#include "gfx/gfxutil.h"
#include "cpluslib/object.h"

class GfxFont;

class GfxCanvas : public Object {
 public:
    virtual ~GfxCanvas();

    /** Draw horizontal line.
        \param pt    origin
        \param npix  number of pixels (width)
        \param color color, in target format.
        \param pat   pattern. Only draw pixels set in this pattern.
        \param alpha alpha value for transparency. */
    virtual void drawHLine(const GfxPoint& pt, int npix, uint32_t color, uint8_t pat, alpha_type alpha) = 0;
    /** Draw vertical line.
        \param pt    origin
        \param npix  number of pixels (height)
        \param color color, in target format.
        \param pat   pattern. Only draw pixels set in this pattern.
        \param alpha alpha value for transparency. */
    virtual void drawVLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha) = 0;
    /** Draw single pixel.
        \param pt    position.
        \param color color, in target format.
        \param alpha alpha value for transparency. */
    virtual void drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha) = 0;
    /** Draw filled rectangle.
        \param rect  area
        \param color color, in target format.
        \param bg    background color, or GFX_TRANSPARENT.
        \param pat   fill pattern.
        \param alpha alpha value for transparency. */
    virtual void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha) = 0;
    /** Copy other surface (pixmap).
        \param pt    anchor point. src's (0,0) is here.
        \param src   input data
        \param rect  display this part of src, relative coordinates.

        `blitSurface(pt, src, GfxRect(src))' will display the whole
        surface. `blitSurface(pt, src, GfxRect(2, 3, 4, 5))' will
        display a 4x5 area from surface src, at (x+2, y+3) up to but not
        including (x+2+4, y+3+5). */
    virtual void blitSurface(const GfxPoint& pt, GfxCanvas& src, GfxRect rect) = 0;
    /** Display pattern (monochrome pixmap).
        \param rect    position, screen coordinates;
        \param pt      position of bit 7, data[0], on the screen;
        \param bytes_per_line  bytes per line in data;
        \param data    pixel data. bit 7 = left, bit 0 = right (VGA bit order);
        \param color   color of "on" bits;
        \param bg      color of "off" bits, or GFX_TRANSPARENT;
        \param alpha   alpha channel.

        \pre pt.x <= rect.x, pt.y <= rect.y

        Note that this function has no constraints on the bytes_per_line
        parameter. bytes_per_line can also be zero or negative for
        interesting effects. */
    virtual void blitPattern(GfxRect rect, const GfxPoint& pt, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha) = 0;
    /** Lock surface. Users must call this before any output
        operation. Recursive locks are permitted. */
    virtual bool lock() = 0;
    /** Unlock surface. Users must call this after any output
        operation. */
    virtual void unlock() = 0;
    /** Update part of output. This will make sure that changes inside
        r will be shown on the screen. Without update, output might or
        might not happen, depending on the type of surface. */
    virtual void update(GfxRect r) = 0;
    /** Compute clipping rectangle. This can be used to optimize
        output. This computes the smallest rectangle p such that
        drawBar(p) and drawBar(r) have the same effect. */
    virtual GfxRect computeClipRect(GfxRect r) = 0;
    /** Get offset transformation to another surface.
        Adjusts #pt such that other.drawPixel(pt) afterwards
        accesses the same pixel as this->drawPixel(pt) before.
        \param other Reference surface
        \param result [in/out] Point to compute
        \return true if transformation can be computed, false otherwise.
        \inv computeOffset(*this,pt) returns true and does not modify pt */
    virtual bool computeOffset(GfxCanvas& other, GfxPoint& pt) = 0;
    /** Read one pixel. \return pixel value in target format. */
    virtual uint32 getPixel(const GfxPoint& pt) = 0;
    /** Get size of this surface */
    virtual GfxPoint getSize() = 0;
    /** Get number of bits used per pixel. */
    virtual int getBitsPerPixel() = 0;

    virtual bool isVisible(GfxRect r);
    virtual bool isClipped(GfxRect r);
};

class GfxSurface : public GfxCanvas {
 protected:
    SDL_Surface* sfc;
    bool owner;
    int locked;

    void setSurface(SDL_Surface* asfc, bool aown = true);

 public:
    GfxSurface(GfxSystemSurface asfc, bool aowner);
    GfxSurface(int wi, int he, GfxSystemPixelFormat fmt);
    ~GfxSurface();

    void drawHLine(const GfxPoint& pt, int npix, uint32_t color, uint8_t pat, alpha_type alpha);
    void drawVLine(const GfxPoint& pt, int npix, uint32_t color, uint8_t pat, alpha_type alpha);
    void drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha);
    void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha);

    void blitSurface(const GfxPoint& pt, GfxCanvas& src, GfxRect rect);
    void blitPattern(GfxRect rect, const GfxPoint& pt, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha);

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxRect computeClipRect(GfxRect r);
    bool computeOffset(GfxCanvas& other, GfxPoint& pt);
    GfxPoint getSize();
    int getBitsPerPixel();

    GfxSystemSurface getSystemSurface();

    uint32 getPixel(const GfxPoint& pt);
};

class GfxFilter : public GfxCanvas {
 protected:
    GfxCanvas& parent;
    GfxFilter(GfxCanvas& parent);
 public:
    GfxCanvas& getParent() const { return parent; }

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxPoint getSize();
    int getBitsPerPixel();

    uint32 getPixel(const GfxPoint& pt);
};

class GfxXorFilter : public GfxCanvas {
    GfxCanvas& parent;
 public:
    GfxXorFilter(GfxCanvas& parent);
    ~GfxXorFilter();

    void drawHLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha);
    void drawVLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha);
    void drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha);
    void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha);

    void blitSurface(const GfxPoint& pt, GfxCanvas& src, GfxRect rect);
    void blitPattern(GfxRect rect, const GfxPoint& pt, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha);

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxRect computeClipRect(GfxRect r);
    bool computeOffset(GfxCanvas& other, GfxPoint& pt);
    GfxPoint getSize();
    int getBitsPerPixel();

    uint32 getPixel(const GfxPoint& pt);
};

class GfxNullCanvas : public GfxCanvas {
 public:
    GfxNullCanvas() { }
    ~GfxNullCanvas();

    void drawHLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha);
    void drawVLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha);
    void drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha);
    void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha);

    void blitSurface(const GfxPoint& pt, GfxCanvas& src, GfxRect rect);
    void blitPattern(GfxRect rect, const GfxPoint& pt, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha);

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxRect computeClipRect(GfxRect r);
    bool computeOffset(GfxCanvas& other, GfxPoint& pt);
    GfxPoint getSize();
    int getBitsPerPixel();

    uint32 getPixel(const GfxPoint& pt);
};

// re-export these functions:
#include "gfx/gfxcompl.h"

#endif
