/**
  *  \file int/if/iterif.cc
  *  \brief Interpreter: Iterator Interface
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "int/if/iterif.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/ion.h"
#include "game/mine.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/ufo.h"
#include "int/args.h"
#include "int/compound.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/value.h"
#include "int/propenum.h"
#include "cpluslib/format.h"
#include "int/contextio.h"
#include "game/fleet.h"
#include "game/objsel.h"

/** Implementation of the "Iterator" function. Iterators implement
    access to a GObjectSelection. Scripts can use this to access a selection's
    current object, and to iterate through the objects. This interface
    is still preliminary.
    - Iterator(n).CurrentIndex: currently active index
    - Iterator(n).Count: number of objects
    - Iterator(n).Id(x): given an index, return that object's Id
    - Iterator(n).Index(id): given an Id, return that object's index; null if none
    - Iterator(n).NextIndex(x,fl): get next index after x. 0 if none found. Flags are "w" to permit wrap, "m" to accept only marked.
    - Iterator(n).PreviousIndex(x,fl): same like NextIndex, but other direction
    - Iterator(n).Object(x): object from index. Still undecided.

    n is:
    - 1,2,3: played ships, planets, bases
    - 10: fleets
    - 21, 22: all ships, planets
    - 30: ufos
    - 31: ion storms
    - 32: minefields */
IntValue*
IFIteratorGet(IntArgBlock& args)
{
    args.checkArgumentCount(1);
    int32_t v;
    if (!checkIntArg(v, args.getNext()))
        return 0;

    if (GObjectType* ty = getObjectTypeFromIteratorId(v)) {
        return new IntIteratorContext(*ty, v);
    } else {
        throw IntError::rangeError();
    }
}

GObjectSelection*
getObjectSelectionFromIteratorId(int32_t id)
{
    switch (id) {
     case 1:  return &current_ship_selection;
     case 2:  return &current_planet_selection;
     case 3:  return &current_base_selection;
     case 10: return &current_fleet_selection;
     case 30: return &current_ufo_selection;
     case 31: return &current_ion_selection;
     case 32: return &current_minefield_selection;
     default: return 0;
    }
}

GObjectType*
getObjectTypeFromIteratorId(int32_t id)
{
    switch (id) {
     case 1:  return &current_ship_type;
     case 2:  return &current_planet_type;
     case 3:  return &current_base_type;
     case 10: return &current_fleet_type;
     case 21: return &current_anyship_type;
     case 22: return &current_anyplanet_type;
     case 30: return &current_ufo_type;
     case 31: return &current_ion_type;
     case 32: return &current_minefield_type;
     default: return 0;
    }
}

namespace {
    enum IntIteratorProperty {
        iitCount,
        iitCurrent,
        iitId,
        iitIndex,
        iitNext,
        iitObject,
        iitPrevious
    };

    class IntIteratorFunction : public IntIndexableValue {
        IntIteratorProperty prop;
        GObjectType& type;
     public:
        IntIteratorFunction(IntIteratorProperty prop, GObjectType& type);

        // IntIndexableValue
        IntValue*   get(IntArgBlock& args);
        void        set(IntArgBlock& args, IntValue* value);
        int32_t     getDimension(int32_t dim);
        IntContext* makeFirstContext();

        // IntValue
        string_t    toString(bool readable);
        IntValue*   clone();
        void        store(IntTagNode&, Stream&);
    };
}

static const TIntMapping iter_map[] = {
    { "COUNT",            iitCount,    0, thInt },
    { "CURRENTINDEX",     iitCurrent,  0, thInt },
    { "ID",               iitId,       0, thArray },
    { "INDEX",            iitIndex,    0, thArray },
    { "NEXTINDEX",        iitNext,     0, thArray },
    { "OBJECT",           iitObject,   0, thArray },
    { "PREVIOUSINDEX",    iitPrevious, 0, thArray },
};

IntIteratorFunction::IntIteratorFunction(IntIteratorProperty prop, GObjectType& type)
    : prop(prop), type(type)
{ }

IntValue*
IntIteratorFunction::get(IntArgBlock& args)
{
    int32_t i, fl;
    switch (prop) {
     case iitId:
        // "Id(index)" => get Id of object
        args.checkArgumentCount(1);
        if (checkIntArg(i, args.getNext(), 0, INT_MAX)) {
            if (type.isValidIndex(i)) {
                return makeIntValue(type.getObjectByIndex(i).getId());
            } else {
                return 0;
            }
        } else {
            return 0;
        }

     case iitIndex:
        // "Index(id)" => find index for Id
        args.checkArgumentCount(1);
        if (checkIntArg(i, args.getNext(), 0, INT_MAX)) {
            for (int index = type.findNextIndex(0); index != 0; index = type.findNextIndex(index))
                if (type.getObjectByIndex(index).getId() == i)
                    return makeIntValue(index);
            return 0;
        } else {
            return 0;
        }

     case iitNext:
        // "Next(index[,flags])" => find next index from index
        args.checkArgumentCount(1, 2);
        if (checkIntArg(i, args.getNext(), 0, INT_MAX)) {
            fl = 0;
            checkFlagArg(fl, 0, args.getNext(), "MW");
            if (fl & 2)
                return makeIntValue(type.findNextIndexWrap(i, fl & 1));
            else
                return makeIntValue(type.findNextIndexNoWrap(i, fl & 1));
        } else {
            return 0;
        }

     case iitPrevious:
        // "Previous(index[,flags])" => find previous index from index
        args.checkArgumentCount(1, 2);
        if (checkIntArg(i, args.getNext(), 0, INT_MAX)) {
            fl = 0;
            checkFlagArg(fl, 0, args.getNext(), "MW");
            if (fl & 2)
                return makeIntValue(type.findPreviousIndexWrap(i, fl & 1));
            else
                return makeIntValue(type.findPreviousIndexNoWrap(i, fl & 1));
        } else {
            return 0;
        }

     case iitObject:
        // FIXME
     default:
        return 0;
    }
}

void
IntIteratorFunction::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

int32_t
IntIteratorFunction::getDimension(int32_t /*dim*/)
{
    return 0;
}

IntContext*
IntIteratorFunction::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue
string_t
IntIteratorFunction::toString(bool /*readable*/)
{
    return "#<array>";
}

IntValue*
IntIteratorFunction::clone()
{
    return new IntIteratorFunction(prop, type);
}

void
IntIteratorFunction::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}




IntIteratorContext::IntIteratorContext(GObjectType& type, uint32_t id)
    : type(type),
      id(id)
{ }

IntIteratorContext::~IntIteratorContext()
{ }

// IntContext:
int32_t
IntIteratorContext::lookup(const string_t& name)
{
    return lookupName(name, iter_map, countof(iter_map));
}

void
IntIteratorContext::set(int32_t index, IntValue* value)
{
    int32_t v;
    switch (index) {
     case iitCurrent:
        if (checkIntArg(v, value, 0, INT_MAX)) {
            GObjectSelection* selection = getObjectSelectionFromIteratorId(id);
            if (selection != 0 && type.isValidIndex(v)) {
                selection->setCurrentIndex(v);
            } else {
                throw IntError::rangeError();
            }
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntIteratorContext::get(int32_t index)
{
    switch (IntIteratorProperty(index)) {
     case iitCount:
        return makeIntValue(type.countObjects());
     case iitCurrent:
        if (GObjectSelection* selection = getObjectSelectionFromIteratorId(id)) {
            return makeIntValue(selection->getCurrentIndex());
        } else {
            return 0;
        }
     case iitId:
     case iitIndex:
     case iitNext:
     case iitObject:
     case iitPrevious:
        return new IntIteratorFunction(IntIteratorProperty(index), type);
     default:
        return 0;
    }
}

IntContext*
IntIteratorContext::clone()
{
    return new IntIteratorContext(type, id);
}

GObject*
IntIteratorContext::getObject()
{
    return 0;
}

string_t
IntIteratorContext::toString(bool readable)
{
    if (readable) {
        return format("Iterator(%d)", id);
    } else {
        return "#<iterator>";
    }
}

void
IntIteratorContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Iterator;
    sv.value = id;
}

void
IntIteratorContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(iter_map, countof(iter_map));
}
