/**
  *  \file int/optimize.cc
  *  \brief Interpreter: Optimizer
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include <cctype>
#include "int/optimize.h"
#include "int/bytecode.h"
#include "int/opcode.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "int/unary.h"
#include "int/value.h"
#include "int/fusion.h"

namespace {
    /** Information about a label. */
    struct IntLabelInfo {
        uint32_t address;       ///< Address of label.
        uint32_t use_count;     ///< Reference count.
        IntLabelInfo()
            : address(0), use_count(0)
            { }
    };

    /** State of optimizer. */
    class IntOptimizerState {
        IntBytecodeObject& bco;
        std::vector<IntLabelInfo> label_info;
        bool had_absolute;
        typedef IntBytecodeObject::pc_t pc_t;

        void initLabelInfo();

        bool doStoreDrop(pc_t pc);
        bool doStoreDropMember(pc_t pc);
        bool doMergeDrop(pc_t pc);
        bool doNullOp(pc_t pc);
        bool doEraseUnusedLabels(pc_t pc);
        bool doInvertJumps(pc_t pc);
        bool doThreadJumps(pc_t pc);
        bool doRemoveUnused(pc_t pc);
        bool doMergeNegation(pc_t pc);
        bool doUnaryCondition(pc_t pc);
        bool doFoldUnaryInt(pc_t pc);
        bool doFoldBinaryInt(pc_t pc);
        bool doFoldJump(pc_t pc);
        bool doPopPush(pc_t pc);
        bool doCompareNC(pc_t pc);

        void clearInstruction(pc_t pc);

     public:
        IntOptimizerState(IntBytecodeObject& bco);
        bool iterate();
    };
}

/*************************** IntOptimizerState ***************************/

/** Constructor. */
IntOptimizerState::IntOptimizerState(IntBytecodeObject& bco)
    : bco(bco), had_absolute(false)
{
    initLabelInfo();
}

/** Initialize label information. Finds all label's addresses and
    reference count. This also checks for absplute jumps. When
    absolute jumps are present, optimisation will not be possible. */
void
IntOptimizerState::initLabelInfo()
{
    label_info.resize(bco.getNumLabels());
    for (IntBytecodeObject::pc_t i = 0; i < bco.getNumInstructions(); ++i) {
        if (bco(i).major == IntOpcode::maJump) {
            if ((bco(i).minor & IntOpcode::jSymbolic) != 0) {
                ASSERT(bco(i).arg < label_info.size());
                if (bco(i).isLabel()) {
                    label_info[bco(i).arg].address = i;
                } else {
                    ++label_info[bco(i).arg].use_count;
                }
            } else {
                had_absolute = true;
            }
        }
    }
}

/** Perform one optimisation iteration.
    \retval true there was a change to the code
    \retval false no possible optimisation found, code unchanged */
bool
IntOptimizerState::iterate()
{
    /* When there are absolute labels, we cannot optimize */
    if (had_absolute)
        return false;

    struct Table {
        bool (IntOptimizerState::*func)(pc_t);           /**< Function to call. */
        IntOpcode::Major          major;                 /**< Required major opcode. */
        pc_t                      observed_insns;        /**< Number of observed instructions after pc. */
        const char*               name;
    };
    static const Table table[] = {
        { &IntOptimizerState::doStoreDrop,         IntOpcode::maStore,    1, "StoreDrop" },
        { &IntOptimizerState::doStoreDropMember,   IntOpcode::maMemref,   1, "StoreDropIM" },
        { &IntOptimizerState::doStoreDropMember,   IntOpcode::maIndirect, 1, "StoreDropIM" },
        { &IntOptimizerState::doMergeDrop,         IntOpcode::maStack,    1, "MergeDrop" },
        { &IntOptimizerState::doNullOp,            IntOpcode::maStack,    0, "NullOp" },
        { &IntOptimizerState::doEraseUnusedLabels, IntOpcode::maJump,     0, "EraseUnusedLabels" },
        { &IntOptimizerState::doInvertJumps,       IntOpcode::maJump,     2, "InvertJumps" },
        { &IntOptimizerState::doThreadJumps,       IntOpcode::maJump,     0, "ThreadJumps" },
        { &IntOptimizerState::doRemoveUnused,      IntOpcode::maJump,     1, "RemoveUnused" },
        { &IntOptimizerState::doRemoveUnused,      IntOpcode::maSpecial,  1, "RemoveUnused" },
        { &IntOptimizerState::doMergeNegation,     IntOpcode::maUnary,    1, "MergeNegation" },
        { &IntOptimizerState::doUnaryCondition,    IntOpcode::maUnary,    1, "UnaryCondition" },
        { &IntOptimizerState::doFoldUnaryInt,      IntOpcode::maPush,     1, "FoldUnaryInt" },
        { &IntOptimizerState::doFoldBinaryInt,     IntOpcode::maPush,     1, "FoldBinaryInt" },
        { &IntOptimizerState::doFoldJump,          IntOpcode::maPush,     1, "FoldJump" },
        { &IntOptimizerState::doPopPush,           IntOpcode::maPop,      1, "PopPush" },
        { &IntOptimizerState::doCompareNC,         IntOpcode::maPush,     1, "CompareNC" },
    };

    bool did = false;
    // printf("iterate():\n");
    for (pc_t i = 0; i < bco.getNumInstructions(); ++i) {
        for (uint32_t ti = 0; ti < countof(table); ++ti) {
            if (bco(i).major == table[ti].major && bco.getNumInstructions() - i > table[ti].observed_insns) {
                if ((this->*table[ti].func)(i)) {
                    // printf("  %s applied at %d\n", table[ti].name, int(i));
                    did = true;
                }
            }
        }
    }

    return did;
}

/** Make instruction at specified pc blank. We use absolute labels as
    null operations. Those will be removed by IntBytecodeObject::compact(). */
void
IntOptimizerState::clearInstruction(pc_t pc)
{
    if (bco(pc).isJumpOrCatch()) {
        /* Jump or catch, so decrement label's reference count */
        --label_info[bco(pc).arg].use_count;
    }
    bco(pc).major = IntOpcode::maJump;
    bco(pc).minor = IntOpcode::jLabel;
    bco(pc).arg   = 0;
}

/************************ Individual Optimisations ***********************/

/** StoreDrop optimisation. Combine STOREx + DROP into POPx. 
    Implemented by decreasing the DROP's counter; if it reaches zero,
    it is removed by doNullOp.

    This pattern appears in assignments. */
bool
IntOptimizerState::doStoreDrop(pc_t pc)
{
    if (bco(pc+1).is(IntOpcode::miStackDrop) && bco(pc+1).arg > 0) {
        bco(pc).major = IntOpcode::maPop;
        --bco(pc+1).arg;
        return true;
    } else {
        return false;
    }
}

/** StoreDrop optimisation for member references. Combine STOREx + DROP into POPx. 
    Implemented by decreasing the DROP's counter; if it reaches zero,
    it is removed by doNullOp.

    This pattern appears in assignments. */
bool
IntOptimizerState::doStoreDropMember(pc_t pc)
{
    if ((bco(pc).minor & IntOpcode::miIMOpMask) == IntOpcode::miIMStore
        && bco(pc+1).is(IntOpcode::miStackDrop)
        && bco(pc+1).arg > 0)
    {
        bco(pc).minor -= IntOpcode::miIMStore;
        bco(pc).minor += IntOpcode::miIMPop;
        --bco(pc+1).arg;
        return true;
    } else {
        return false;
    }
}

/** MergeDrop optimisation.
    Combine two DROP into one.

    This pattern appears in "For ... Select Case ... Break", which
    generates one DROP for the selector, and one for the loop
    bound. */
bool
IntOptimizerState::doMergeDrop(pc_t pc)
{
    if (bco(pc).is(IntOpcode::miStackDrop) && bco(pc+1).is(IntOpcode::miStackDrop)) {
        bco(pc+1).arg += bco(pc).arg;
        clearInstruction(pc);
        return true;
    } else {
        return false;
    }
}

/** Remove null operations. Those are DROP 0 and SWAP 0.
    DROP 0 appears after doStoreDrop(). */
bool
IntOptimizerState::doNullOp(pc_t pc)
{
    if ((bco(pc).is(IntOpcode::miStackDrop) || bco(pc).is(IntOpcode::miStackSwap)) && bco(pc).arg == 0) {
        clearInstruction(pc);
        return true;
    } else {
        return false;
    }
}

/** Erase unused labels. Those appear frequently. */
bool
IntOptimizerState::doEraseUnusedLabels(pc_t pc)
{
    /* Remove unreferenced labels */
    if (bco(pc).isLabel() && (bco(pc).minor & IntOpcode::jSymbolic) != 0 && label_info[bco(pc).arg].use_count == 0) {
        clearInstruction(pc);
        return true;
    } else {
        return false;
    }
}

/** Invert jumps. Converts a conditional jump across another jump into
    a single conditional jump.

    This pattern appears frequently in conditions. */
bool
IntOptimizerState::doInvertJumps(pc_t pc)
{
    if (bco(pc).isRegularJump()
        && label_info[bco(pc).arg].address == pc+2
        && bco(pc+1).isRegularJump()
        && (bco(pc+1).minor & IntOpcode::jPopAlways) == 0)
    {
        int nextMinor = (bco(pc+1).minor & ~(bco(pc).minor & IntOpcode::jAlways)) | (bco(pc).minor & IntOpcode::jPopAlways);
        if ((nextMinor & IntOpcode::jAlways) == 0) {
            // Second jump is never taken. Eliminate both.
            clearInstruction(pc);
            clearInstruction(pc+1);
            if (nextMinor & IntOpcode::jPopAlways) {                
                bco(pc+1).major = IntOpcode::maStack;
                bco(pc+1).minor = IntOpcode::miStackDrop;
                bco(pc+1).arg = 1;
            }
        } else {
            // Second jump is taken sometimes. Eliminate first jump.
            clearInstruction(pc);
            bco(pc+1).minor = nextMinor;
        }
        return true;
    } else {
        return false;
    }
}

/** Thread jumps. When a jump targets a label or another jump, adjust
    its target to minimize work at runtime.
    - skip over labels, so that if several jumps target these labels, all jump to the last label,
      and the others are unreferenced and can be removed
    - follow unconditional jumps and directly jump to their targets
    - follow at most one backward jump, to avoid chasing infinite loops

    This pattern appears frequently. */
bool
IntOptimizerState::doThreadJumps(pc_t pc)
{
    /* Is it actually a jump? */
    if (!bco(pc).isRegularJump())
        return false;

    bool had_backward_jump = false;
    uint16_t target_label = bco(pc).arg;
    while (1) {
        pc_t target_address = label_info[target_label].address;
        ASSERT(bco(target_address).isLabel());
        if (target_address+1 >= bco.getNumInstructions()) {
            /* Jump to label at end of routine */
            break;
        } else if (bco(target_address+1).isLabel() && (bco(target_address+1).minor & IntOpcode::jSymbolic) != 0) {
            /* Jump to a label, which is followed by a label. Jump to the later label,
               to make the former unreferenced and removable. */
            target_label = bco(target_address+1).arg;
            ASSERT(label_info[target_label].address == target_address+1);
        } else if (bco(target_address+1).isRegularJump()
                   && (bco(target_address+1).minor == (IntOpcode::jAlways | IntOpcode::jSymbolic)
                       || ((bco(pc).minor & IntOpcode::jPopAlways) == 0
                           && (bco(target_address+1).minor & IntOpcode::jPopAlways) == 0
                           && (bco(pc).minor & ~bco(target_address+1).minor) == 0)))
        {
            /* Jump to an unconditional jump (frequent in if-within-loops),
               or to a conditional jump with the same condition (in a:=b xor c). */
            pc_t arg = bco(target_address+1).arg;
            if (label_info[arg].address <= target_address) {
                /* At most one backward jump per iteration to avoid infinite loops */
                if (had_backward_jump)
                    break;
                had_backward_jump = true;
            }
            target_label = arg;
        } else {
            /* Not a jump or label */
            break;
        }
    }

    /* Did we change anything? */
    if (label_info[target_label].address == pc+1) {
        /* It is a jump to the next instruction. Delete it. */
        if (bco(pc).minor & IntOpcode::jPopAlways) {
            clearInstruction(pc);
            bco(pc).major = IntOpcode::maStack;
            bco(pc).minor = IntOpcode::miStackDrop;
            bco(pc).arg = 1;
        } else {
            clearInstruction(pc);
        }
        return true;
    } else if (target_label != bco(pc).arg) {
        /* We followed a jump chain. */
        --label_info[bco(pc).arg].use_count;
        bco(pc).arg = target_label;
        ++label_info[bco(pc).arg].use_count;
        return true;
    } else {
        return false;
    }
}

/** Remove unused code. Removes code following an unconditional jump,
    a UTHROW, or SRETURN instruction. Such code is never executed.

    This appears in code such as "If a Then Return Else ..." */
bool
IntOptimizerState::doRemoveUnused(pc_t pc)
{
    if ((bco(pc).isRegularJump() && (bco(pc).minor & IntOpcode::jAlways) == IntOpcode::jAlways)
        || (bco(pc).is(IntOpcode::miSpecialThrow))
        || (bco(pc).is(IntOpcode::miSpecialTerminate))
        || (bco(pc).is(IntOpcode::miSpecialReturn)))
    {
        // Jump, throw or return
        pc_t i = pc+1;
        while (i < bco.getNumInstructions() && !bco(i).isLabel()) {
            clearInstruction(i);
            ++i;
        }
        return i > pc+1;
    } else {
        return false;
    }
}

/** MergeNegation optimisation. Merges two unary logic/sign instructions
    into one if possible.

    This appears in "If IsEmpty(Zap(x))" (UZAP + UISEMPTY => UNOT2),
    or in "For i:=-1 To +1" (UNEG from boundary + UPOS used as type-check),
    and in explicitly written code ("Zap(Zap(x))"). */
bool
IntOptimizerState::doMergeNegation(pc_t pc)
{
    /* Merge pairs of unary logic into single instructions */
    enum { None = -1, RepFalse = -2, ZapBool = -3 };
    int result = None;

    if (!bco(pc+1).is(IntOpcode::maUnary))
        return false;

    switch (bco(pc).minor) {
     case unNot:
        switch (bco(pc+1).minor) {
         case unNot:     result = unBool;    break;
         case unBool:    result = unNot;     break;
         case unIsEmpty: result = unIsEmpty; break;
        }
        break;
     case unBool:
        switch (bco(pc+1).minor) {
         case unNot:     result = unNot;     break;
         case unBool:    result = unBool;    break;
         case unNot2:    result = unNot2;    break;
         case unIsEmpty: result = unIsEmpty; break;
         case unZap:     result = ZapBool;   break;
        }
        break;
     case unNot2:
        switch (bco(pc+1).minor) {
         case unBool:    result = unNot2;    break;
         case unIsEmpty: result = RepFalse;  break;
        }
        break;
     case unIsEmpty:
        switch (bco(pc+1).minor) {
         case unBool:    result = unIsEmpty; break;
         case unIsEmpty: result = RepFalse;  break;
        }
        break;
     case unZap:
        switch (bco(pc+1).minor) {
         case unNot2:    result = unNot2;    break;
         case unIsEmpty: result = unNot2;    break;
         case unZap:     result = unZap;     break;
        }
        break;
     case unNeg:
        switch (bco(pc+1).minor) {
         case unNeg:     result = unPos;     break;
         case unPos:     result = unNeg;     break;
        }
        break;
     case unPos:
        switch (bco(pc+1).minor) {
         case unNeg:     result = unNeg;     break;
         case unPos:     result = unPos;     break;
         case unInc:     result = unInc;     break;
         case unDec:     result = unDec;     break;
        }
        break;
     case unInc:
        switch (bco(pc+1).minor) {
         case unDec:     result = unPos;     break;
         case unPos:     result = unInc;     break;
        }
        break;
     case unDec:
        switch (bco(pc+1).minor) {
         case unInc:     result = unPos;     break;
         case unPos:     result = unDec;     break;
        }
        break;
    }

    if (result == None) {
        /* Cannot reduce this one */
        return false;
    } else if (result == RepFalse) {
        /* Replace by 'drop 1; pushbool 0' */
        bco(pc).major   = IntOpcode::maStack;
        bco(pc).minor   = IntOpcode::miStackDrop;
        bco(pc).arg     = 1;
        bco(pc+1).major = IntOpcode::maPush;
        bco(pc+1).minor = IntOpcode::sBoolean;
        bco(pc+1).arg   = 0;
        return true;
    } else if (result == ZapBool) {
        /* Replace by 'uzap; ubool' (potentially fewer temporaries) */
        bco(pc).minor = unZap;
        bco(pc+1).minor = unBool;
        return true;
    } else {
        /* Replace by 'u<whatever>' */
        clearInstruction(pc);
        bco(pc+1).minor = result;
        return true;
    }
}

/** UnaryCondition optimisation. If a logic instruction is followed by
    a conditional jump, modify the jump's condition to evaluate the
    logic, e.g. merge UNOT+JTP intp JFP.

    This appears often in conditions such as "If IsEmpty(x)". */
bool
IntOptimizerState::doUnaryCondition(pc_t pc)
{
    /* u<logic> / j<cc>p -> j<cc'>p */
    if (bco(pc+1).isRegularJump() && (bco(pc+1).minor & IntOpcode::jPopAlways) != 0) {
        int oldCond = bco(pc+1).minor;
        int newCond = 0;
        if (bco(pc).is(unIsEmpty)) {
            /* uisempty: t->e, f->tf, e->never */
            if (oldCond & IntOpcode::jIfTrue)
                newCond |= IntOpcode::jIfEmpty;
            if (oldCond & IntOpcode::jIfFalse)
                newCond |= IntOpcode::jIfTrue | IntOpcode::jIfFalse;
        } else if (bco(pc).is(unNot)) {
            /* unot: t->f, f->t, e->e */
            if (oldCond & IntOpcode::jIfTrue)
                newCond |= IntOpcode::jIfFalse;
            if (oldCond & IntOpcode::jIfFalse)
                newCond |= IntOpcode::jIfTrue;
            if (oldCond & IntOpcode::jIfEmpty)
                newCond |= IntOpcode::jIfEmpty;
        } else if (bco(pc).is(unZap)) {
            /* uzap: t->t, f->never, e->fe */
            if (oldCond & IntOpcode::jIfTrue)
                newCond |= IntOpcode::jIfTrue;
            if (oldCond & IntOpcode::jIfEmpty)
                newCond |= IntOpcode::jIfEmpty | IntOpcode::jIfFalse;
        } else if (bco(pc).is(unNot2)) {
            /* unot2: t->fe, f->t, e->never */
            if (oldCond & IntOpcode::jIfTrue)
                newCond |= IntOpcode::jIfFalse | IntOpcode::jIfEmpty;
            if (oldCond & IntOpcode::jIfFalse)
                newCond |= IntOpcode::jIfTrue;
        } else if (bco(pc).is(unBool)) {
            /* ubool / jccp --> jccp */
            newCond = oldCond;
        } else {
            return false;
        }

        /* Update it */
        clearInstruction(pc);
        if (newCond == 0) {
            /* This will become a "jump never" */
            clearInstruction(pc+1);
            bco(pc+1).major = IntOpcode::maStack;
            bco(pc+1).minor = IntOpcode::miStackDrop;
            bco(pc+1).arg   = 1;
        } else {
            /* Update jump */
            bco(pc+1).minor = newCond | IntOpcode::jPopAlways | IntOpcode::jSymbolic;
        }
        return true;
    } else {
        return false;
    }
}

/** Constant-fold unary operations on integer constants. The most
    frequent occurrences are negative literals (encoded as "pushint,
    uneg"), "For i:=1 to 10" type checks ("pushint 1, upos"), and
    "Z(0)" ("pushint 0, uzap"). The others are included for
    completeness. */
bool
IntOptimizerState::doFoldUnaryInt(pc_t pc)
{
    if ((bco(pc).minor != IntOpcode::sInteger && bco(pc).minor != IntOpcode::sBoolean) || !bco(pc+1).is(IntOpcode::maUnary))
        return false;

    switch (bco(pc+1).minor) {
     case unZap:
     case unNeg:
     case unPos:
     case unNot:
     case unNot2:
     case unBool:
     case unAbs:
     case unIsEmpty:
     case unIsString:
     case unIsNum:
     case unTrunc:
     case unRound:
     case unInc:
     case unDec:
     case unBitNot:
        {
            IntValue* result;
            if (bco(pc).minor == IntOpcode::sInteger) {
                IntIntIntValue iv(+int16_t(bco(pc).arg));
                result = executeUnaryOperation(bco(pc+1).minor, &iv);
            } else if (int16_t(bco(pc).arg) < 0) {
                result = executeUnaryOperation(bco(pc+1).minor, 0);
            } else {
                IntIntBoolValue iv(+int16_t(bco(pc).arg));
                result = executeUnaryOperation(bco(pc+1).minor, &iv);
            }

            /* Can we encode the result? */
            bool did;
            if (result == 0) {
                bco(pc).minor = IntOpcode::sBoolean;
                bco(pc).arg   = uint16_t(-1U);
                did = true;
            } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(result)) {
                if (iv->isBool()) {
                    bco(pc).minor = IntOpcode::sBoolean;
                    bco(pc).arg   = (iv->getValue() != 0);
                    did = true;
                } else if (iv->getValue() >= -32767 && iv->getValue() <= +32767) {
                    bco(pc).minor = IntOpcode::sInteger;
                    bco(pc).arg   = iv->getValue();
                    did = true;
                } else {
                    did = false;
                }
            } else {
                did = false;
            }
            delete result;
            if (did)
                clearInstruction(pc+1);
            return did;
        }
     default:
        return false;
    }
}

/** Convert binary operations using a single constant integer
    parameter into unary operations.

    The "+1"/"-1" case appears frequently in user code, the others are
    included for completeness. */
bool
IntOptimizerState::doFoldBinaryInt(pc_t pc)
{
    if (bco(pc).minor != IntOpcode::sInteger || !bco(pc+1).is(IntOpcode::maBinary))
        return false;

    int16_t value = bco(pc).arg;
    switch (bco(pc+1).minor) {
     case biAdd:
        /* "+ 1" => uinc, "+ -1" => udec, "+0" => upos */
        if (value == 0 || value == 1 || value == -1) {
            bco(pc+1).major = IntOpcode::maUnary;
            bco(pc+1).minor = (value == 0 ? unPos : value == 1 ? unInc : unDec);
            clearInstruction(pc);
            return true;
        } else {
            return false;
        }
     case biSub:
        /* "- +1" => uinc, "- -1" => uinc, "+0" => upos */
        if (value == 0 || value == 1 || value == -1) {
            bco(pc+1).major = IntOpcode::maUnary;
            bco(pc+1).minor = (value == 0 ? unPos : value == 1 ? unDec : unInc);
            clearInstruction(pc);
            return true;
        } else {
            return false;
        }
     case biMult:
     case biDivide:
     case biIntegerDivide:
        /* "* +1" => upos, "* -1" => uneg */
        if (value == 1 || value == -1) {
            bco(pc+1).major = IntOpcode::maUnary;
            bco(pc+1).minor = (value == 1 ? unPos : unNeg);
            clearInstruction(pc);
            return true;
        } else {
            return false;
        }
     case biPow:
        /* "^ +1" => upos */
        if (value == 1) {
            bco(pc+1).major = IntOpcode::maUnary;
            bco(pc+1).minor = unPos;
            clearInstruction(pc);
            return true;
        } else {
            return false;
        }
     default:
        return false;
    }
}

/** Fold conditional jump on constant. If a literal is immediately
    followed by a conditional jump, or by a jump to a conditional
    jump, evaluate that condition and generate an unconditional jump
    instead.

    This appears in 'Do While True' or 'FindShip(1)'. */
bool
IntOptimizerState::doFoldJump(pc_t pc)
{
    if (bco(pc).minor != IntOpcode::sInteger && bco(pc).minor != IntOpcode::sBoolean)
        return false;
    if (!bco(pc+1).isRegularJump())
        return false;

    /* Figure out condition */
    int cond;
    if (bco(pc).minor == IntOpcode::sBoolean && int16_t(bco(pc).arg) < 0)
        cond = IntOpcode::jIfEmpty;
    else if (bco(pc).arg == 0)
        cond = IntOpcode::jIfFalse;
    else
        cond = IntOpcode::jIfTrue;

    /* Check jump */
    if ((bco(pc+1).minor & IntOpcode::jAlways) == IntOpcode::jAlways) {
        /* Unconditional jump. Check whether it jumps at a conditional jump. */
        if (bco(pc+1).minor & IntOpcode::jPopAlways) {
            /* Pathological case: push + jp, turn into j */
            bco(pc+1).minor &= ~IntOpcode::jPopAlways;
            clearInstruction(pc);
            return true;
        }

        /* Follow this jump */
        uint16_t target_label = bco(pc+1).arg;
        pc_t target_address = label_info[target_label].address;
        if (target_address+1 >= bco.getNumInstructions() || !bco(target_address+1).isRegularJump()) {
            /* Jump to end of routine or not to a jump; keep it */
            return false;
        }

        if (bco(target_address+1).minor & cond) {
            /* Jump to conditional jump that will be taken; follow it */
            if (bco(target_address+1).minor & IntOpcode::jPopAlways) {
                clearInstruction(pc);
            }
            --label_info[bco(pc+1).arg].use_count;
            bco(pc+1).arg = bco(target_address+1).arg;
            ++label_info[bco(pc+1).arg].use_count;
            return true;
        } else {
            /* Conditional jump is not taken; we cannot optimize this */
            return false;
        }
    } else {
        /* Conditional jump */
        if (bco(pc+1).minor & IntOpcode::jPopAlways) {
            /* This jump will pop, drop the push */
            clearInstruction(pc);
        }

        if (bco(pc+1).minor & cond) {
            /* This jump will be taken. Make it unconditional. */
            bco(pc+1).minor |= IntOpcode::jAlways;
            bco(pc+1).minor &= ~IntOpcode::jPopAlways;
        } else {
            /* This jump will not be taken. Drop it. */
            clearInstruction(pc+1);
        }
    }

    return true;
}

/** Fold pop followed by push into store. This appears when a variable is
    set and then immediately used, as in 'arg := expr; foo(arg)'. */
bool
IntOptimizerState::doPopPush(pc_t pc)
{
    /* Next instruction must be push with same address, but may not be
       a named variable: assigning those implies a type-cast. */
    if (bco(pc+1).major == IntOpcode::maPush
        && bco(pc).minor == bco(pc+1).minor
        && bco(pc).arg == bco(pc+1).arg
        && bco(pc).minor != IntOpcode::sNamedVariable)
    {
        /* do it */
        bco(pc+1).major = IntOpcode::maStore;
        clearInstruction(pc);
        return true;
    } else {
        /* not applicable */
        return false;
    }
}

/** Convert push followed by case-blind comparison into regular
    comparisons if it can be proven that case-blindness is not needed. 
    This removes the need for upcasing in the comparison. The same
    applies for First/Rest/FindStr. Note that only equality
    comparisons can use this optimisation; ordering comparisons depend
    on upcasing (i.e. "a" < "[" in case-blind mode, "a" > "[" in
    case-sensitive mode). */
bool
IntOptimizerState::doCompareNC(pc_t pc)
{
    const uint8_t pm = bco(pc).minor;
    const uint8_t cm = bco(pc+1).minor;

    /* First must be push literal */
    if (pm != IntOpcode::sInteger && pm != IntOpcode::sBoolean && pm != IntOpcode::sLiteral)
        return false;

    /* Second must be case-blind compare */
    if (bco(pc+1).major != IntOpcode::maBinary)
        return false;
    if (cm != biCompareEQ_NC && cm != biCompareNE_NC && cm != biFirstStr_NC && cm != biRestStr_NC && cm != biFindStr_NC)
        return false;

    /* If first is pushlit, it must be an acceptable literal */
    if (pm == IntOpcode::sLiteral) {
        IntValue* iv = bco.getLiteral(bco(pc).arg);
        if (dynamic_cast<IntIntValue*>(iv) != 0 || dynamic_cast<IntFloatValue*>(iv) != 0) {
            /* accept */ ;
        } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(iv)) {
            /* accept if valid */
            using std::isalpha;
            string_t v = sv->getValue();
            for (string_t::size_type i = 0; i < v.size(); ++i)
                if (isalpha(uint8_t(v[i])))
                    return false;
        } else {
            /* reject */
            return false;
        }
    }

    /* OK */
    --bco(pc+1).minor;
    return true;
}

/** Optimize the given bytecode object. It must not have been relocated yet.
    \param bco [in/out] Bytecode object
    \param level [in] Optimisation level to apply */
void
optimize(IntBytecodeObject& bco, int /*level*/)
{
    unfuseInstructions(bco);
    while (IntOptimizerState(bco).iterate())
        bco.compact();
    fuseInstructions(bco);
}
