/**
  *  \file misc.cc
  *  \brief Miscellaneous utilities
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <cstdlib>
#include <cerrno>
#include <climits>
#include <cstring>
#include "util/misc.h"
#include "game/pref.h"

using namespace std;  /* to access unqualified isXXX/toXXX, some libc++'s
                         still have these as macros */

/** String Match, PHost way.
    \param pattern  Pattern we want to match against. Consists of
                    an initial sequence of capital letters, followed
                    by a sequence of other letters. Lower-case letters
                    are optional, the rest is mandatory.
    \param tester   The string to test against the pattern.

    Matching is case-insensitive.

    Example: stringMatch("ENglish", p) matches if p is
    "english", or any abbreviation of "english" up to "en". */
bool
stringMatch(const char* pattern, const char* tester)
{
    while (*pattern) {
        unsigned char p = *pattern++;
        if (!*tester) {
            /* tester ends before pattern. Permitted? */
            return islower(p);
        } else {
            /* character must match exactly otherwise */
            if (tolower(p) != tolower((unsigned char) *tester))
                return false;
        }
        ++tester;
    }
    /* pattern ends -- tester must end, too */
    return *tester == 0;
}

bool
stringValue(string_t s, long& val)
{
    string_t::size_type pos;
    return stringValue(s, val, pos);
}

/** Parse string into integer.
    \param s     (in) Input
    \param val   (out) Value, only modified if s is completely valid
    \param pos   (out) Number of valid characters, if string is partially
                 valid
    \returns true if string is completely valid; false if not. If the
    string is not completely valid, pos is set to the number of valid
    characters */
bool
stringValue(string_t s, long& val, string_t::size_type& pos)
{
    const char* c = s.c_str();
    char* end;
    
    errno = 0;
    long v = std::strtol(c, &end, 10);
    if (errno == ERANGE) {
        pos = 0;
        return false;
    }
    while (*end && isspace(static_cast<unsigned char>(*end)))
        ++end;
    if (!*c || *end) {
        pos = end - c;
        return false;
    }
    pos = 0;
    val = v;
    return true;
}

bool
stringValue(string_t s, int& val)
{
    string_t::size_type pos;
    return stringValue(s, val, pos);
}

bool
stringValue(string_t s, int& val, string_t::size_type& pos)
{
    long l;
    if (!stringValue(s, l, pos))
        return false;
    if (l < INT_MIN || l > INT_MAX) {
        pos = 0;
        return false;
    }
    val = l;
    return true;
}

/** Parse a flag list.
    \param s       string we got from user
    \param data    allowed options, in caps (e.g., for fcodes, "SPBRAU")
    \param value   pointer to integer variable with result (first option
                   has value 1, second has value 2, third has 4, etc.)
    \param races   pointer to player set. Can be NULL, then
                   we'll not accept races. *races can be preinitialized
                   to a default value (*value can not). Note that the result
                   may contain bits outside 1..11.
    \param accept_bad_flags if set, accept unknown flags.
    \returns true on success, false on error */
bool
parseFlags(const string_t& s, const char* data, int* value,
           SmallSet<int>* races, bool accept_bad_flags)
{
    // FIXME: /races/ cannot currently be GPlayerSet because that is defined in game/, not util/.
    static const char eleven[] = "123456789AB";
    string_t::size_type t = 0;
    
    *value = 0;
    while (t < s.length()) {
        unsigned char c = s[t];
        if (races && (c == '+' || c == '-')) {
            bool negate = (c == '-');
            *races = SmallSet<int>();
            while (++t < s.length()) {
                c = s[t];
                const char* p = strchr(eleven, toupper(c));
                if (!p)
                    return false;
                *races |= p - eleven + 1;
            }
            if (negate)
                *races ^= SmallSet<int>::fromInteger(~0UL);
            break;
        } else {
            const char* p = strchr(data, toupper(c));
            if (p)
                *value |= 1 << (p - data);
            else if (!accept_bad_flags && !isspace(c))
                return false;
            ++t;
        }
    }
    return true;
}

/** Parse a range. Syntax is one of
    - "nn", means set min=max=nn
    - "nn-", means set min=nn, do not modify max
    - "nn-mm", means set min=nn, max=mm

    Note that this function does not enforce that the result actually is a real range
    with min<=max, and does not enforce that the returned range is a subrange of [min,max].
    It just parses the numbers, and min/max are only default values.
    
    \param s    [in] User input
    \param min  [in/out] Default range start
    \param max  [in/out] Default range end
    \param pos  [out] Position where parsing stopped if return value is false
    \return true iff input was completely valid, false otherwise (min,max unmodified, pos set) */
bool
parseRange(const string_t& s, int& min, int& max, string_t::size_type& pos)
{
    int min_out, max_out;

    string_t::size_type p = s.find('-');
    if (p != s.npos) {
        // two parts
        if (!stringValue(s.substr(0, p), min_out, pos))
            return false;

        // do we have a maximum?
        ++p;
        while (p < s.size() && isspace(static_cast<unsigned char>(s[p])))
            ++p;

        if (p < s.size()) {
            // yes, case three
            if (!stringValue(s.substr(p), max_out, pos)) {
                pos += p;
                return false;
            }
            min = min_out;
            max = max_out;
            return true;
        } else {
            // no, case two
            min = min_out;
            return true;
        }
    } else {
        // one part, case one
        if (!stringValue(s, min_out, pos))
            return false;
        min = max = min_out;
        return true;
    }
}

/** Convert number to string. */
string_t
numToString(int32 number)
{
    string_t result = itoa(number);
    if (getUserPreferences().DisplayThousandSep()) {
        // The limit is to avoid placing a thousands-separator as "-,234"
        size_t i = result.size();
        size_t limit = (i > 0 && (result[0] < '0' || result[0] > '9')) ? 4 : 3;
        while (i > limit) {
            i -= 3;
            result.insert(i, ",");
        }
    }
    return result;
}

/** Convert clan count to string. */
string_t
clansToString(int32 number)
{
    return getUserPreferences().DisplayClans()
        ? numToString(number) + "c"
        : numToString(100*number);
}

/** Convert number to roman numeral. */
string_t
numToRoman(int32 number)
{
    /* range check */
    if (number <= 0 || number > 4999)
        return numToString(number);

    /* thousands */
    string_t result;
    while (number >= 1000) {
        result += 'M';
        number -= 1000;
    }

    /* hundreds, tens, ones */
    static const char seq[] = { 'M', 'D', 'C', 'L', 'X', 'V', 'I' };
    static const int  val[] = { 100, 10, 1 };

    for (int i = 0; i < 3; ++i) {
        int digit = number / val[i];
        number %= val[i];
        if (digit == 9) {
            /* CM, XC, IX */
            result += seq[2*i+2];
            result += seq[2*i];
        } else if (digit >= 5) {
            /* DC, LX, VI */
            result += seq[2*i+1];
            result.append(digit-5, seq[2*i+2]);
        } else if (digit == 4) {
            /* CD, XL, IV */
            result += seq[2*i+2];
            result += seq[2*i+1];
        } else {
            /* C, X, I */
            result.append(digit, seq[2*i+2]);
        }
    }

    return result;
}
