/**
  *  \file util/recode.cc
  *  \brief Character-set Recoding
  *
  *  Because Planets is being played on a number of platforms, we must
  *  occasionally recode some strings.
  *
  *  PCC2 now internally uses UTF-8 exclusively. This makes it easy to
  *  interface to different system character sets (UCS-2 on Windows,
  *  locale-specific on Unix). We must, however, make sure that
  *  characters we intend to store in game files are actually
  *  encodable (i.e. in a game using a cyrillic character set you
  *  cannot name a ship using LATIN CAPITAL LETTER A WITH DIAERESIS,
  *  but in a game using Latin-1 or codepage-437, you can, regardless
  *  of your actual system locale).
  *
  *  Strings read from game files are translated from their codepage
  *  into UTF-8, and back when saving. We do not support multibyte
  *  encodings in game files.
  *
  *  We assume that all character sets have the standard ASCII set at
  *  0..127.
  *
  *  PCC2 used to use a pair of codepages here, an "external" one for
  *  game characters, and an "internal" one for fonts, file names and
  *  everything inside, which has the drawback of forcing players to
  *  use their file name character set if they want to see their file
  *  names (and get proper fonts!), and being unable to support
  *  Unicode locales.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "util/recode.h"
#include "cpluslib/misc.h"
#include "cpluslib/string.h"
#include "cpluslib/assert.h"
#include "cpluslib/types.h"
#include "cpluslib/utf8.h"

/*********************** Character Set Definitions ***********************/

struct CharacterSet::Data {
    const char*  name;
    const char*  full_name;
    uint16       chars[128];
};

namespace {
    struct AliasDefinition {
        const char* alias_name;
        int index;
    };
    struct MnemonicDefinition {
        uint16_t code;
        char rfc[4];
    };
}

/** Character set definitions.
    Character set definitions were generated using a small Perl
    script from the mapping tables on unicode.org.

    We only store the "high" half of each character set here, the
    lower half is assumed to be plain 7-bit ASCII. For each of the
    high 128 characters, we store the unicode value.

    An additional mapping from Unicode to RFC 1345's short mnemonics
    makes it simple to generate replacements for nonexact matches, and
    probably other interesting things in the future. */
#include "charsets/charsets.inc"

/************************* Character Set Enquiry *************************/

/** Get a character set by name.
    \returns character set, null if no match. */
CharacterSet
CharacterSet::getCharsetByName(string_t name)
{
    // Remove dashes
    string_t::size_type n;
    while ((n = name.find('-')) != name.npos)
        name.erase(n, 1);

    // Check charsets
    for (std::size_t n = 0; n < countof(::charsets); ++n)
        if (strCaseCmp(name, charsets[n].name) == 0)
            return CharacterSet(&charsets[n]);

    // Check aliases
    for (std::size_t n = 0; n < countof(::alias); ++n)
        if (strCaseCmp(name, alias[n].alias_name) == 0)
            return CharacterSet(&charsets[alias[n].index]);
    return CharacterSet(0);
}

/** Check whether character set name is UTF-8. Note that UTF-8 cannot
    be represented by a CharacterSet object; this is intended only to
    support a consistent naming of UTF-8 everywhere a character set
    including UTF-8 is required. */
bool
CharacterSet::isCharsetNameUtf8(string_t name)
{
    return strCaseCmp(name, "utf-8") == 0
        || strCaseCmp(name, "utf8") == 0;
}

/** Get number of supported character sets. */
std::size_t
CharacterSet::getNumCharsets()
{
    return countof(::charsets);
}

/** Get a character set by index.
    \pre n \in [0, getNumCharsets()) */
CharacterSet
CharacterSet::getCharset(std::size_t n)
{
    ASSERT(n < countof(::charsets));
    return CharacterSet(&charsets[n]);
}

/****************************** CharacterSet *****************************/

/** Get short name of character set. This is what is in config files
    etc. */
const char*
CharacterSet::getName() const
{
    return data ? data->name : "default";
}

/** Get description. This is a one-liner that can be shown in
    menus. */
const char*
CharacterSet::getDescription() const
{
    // FIXME: i18n!
    return data ? data->full_name : "default";
}

/** Return Unicode value of a character. */
uint16
CharacterSet::getUnicode(char c) const
{
    unsigned char uc = c;
    return (data && uc >= 128) ? data->chars[uc-128] : uc;
}

/** Return character corresponding to a Unicode character.
    \returns 0 if mapping fails. */
char
CharacterSet::getCharacter(uint16 uni) const
{
    if (uni < 128)
        return uni;
    if (!data) {
        if (uni < 256)
            return uni;
    } else {
        for (unsigned i = 0; i < 128; ++i)
            if (data->chars[i] == uni)
                return i+128;
    }
    return 0;
}

/** Convert string using this character set to UTF-8. */
string_t
CharacterSet::convertToUtf8(const string_t str) const
{
    string_t result;
    for (string_t::size_type i = 0; i < str.size(); ++i)
        appendUtf8(result, getUnicode(str[i]));
    return result;
}

/** Convert UTF-8 to this character set. Unencodable characters will
    be dropped. */
string_t
CharacterSet::convertFromUtf8(const string_t str) const
{
    string_t result;
    Utf8Reader rdr(str.data(), str.size());
    while (rdr.hasMore()) {
        char c = getCharacter(rdr.get());
        if (c)
            result += c;
    }
    return result;
}

/*********************** Global Game Character Set ***********************/

static CharacterSet game_character_set;

/** Set game character set. */
void
setGameCharacterSet(CharacterSet gcs)
{
    game_character_set = gcs;
}

/** Get currently-used game character set. */
CharacterSet
getGameCharacterSet()
{
    return game_character_set;
}

/** Check whether Unicode character can be encoded in game character set. */
bool
isReliableGameCharacter(uint32_t ch)
{
    return game_character_set.getCharacter(ch) != 0;
}

/** Convert UTF-8 string to game character set. */
string_t
convertUtf8ToGame(string_t str)
{
    return game_character_set.convertFromUtf8(str);
}

/** Convert game character string to UTF-8. */
string_t
convertGameToUtf8(string_t str)
{
    return game_character_set.convertToUtf8(str);
}
